@@ -243,9 +243,8 @@ addons_notify_load_status_cb (MidoriView*      view,
                               MidoriExtension* extension)
 {
     const gchar* uri = midori_view_get_display_uri (view);
-    WebKitWebView* web_view = WEBKIT_WEB_VIEW (midori_view_get_web_view (view));
 
-    if (webkit_web_view_get_view_source_mode (web_view))
+    if (midori_tab_get_view_source (MIDORI_TAB (view)))
         return;
 
     if (uri && *uri)
@@ -482,12 +481,12 @@ addons_open_in_editor_clicked_cb (GtkWidget* toolitem,
 
         g_object_get (settings, "text-editor", &text_editor, NULL);
         if (text_editor && *text_editor)
-            sokoke_spawn_program (text_editor, TRUE, element->fullpath, TRUE);
+            sokoke_spawn_program (text_editor, TRUE, element->fullpath, TRUE, FALSE);
         else
         {
             gchar* element_uri = g_filename_to_uri (element->fullpath, NULL, NULL);
-            sokoke_show_uri (NULL, element_uri,
-                             gtk_get_current_event_time (), NULL);
+            gboolean handled = FALSE;
+            g_signal_emit_by_name (midori_browser_get_current_tab (browser), "open-uri", element_uri, &handled);
             g_free (element_uri);
         }
 
@@ -523,8 +522,9 @@ addons_open_target_folder_clicked_cb (GtkWidget* toolitem,
     folder_uri = g_filename_to_uri (folder, NULL, NULL);
     g_free (folder);
 
-    sokoke_show_uri (gtk_widget_get_screen (GTK_WIDGET (addons->treeview)),
-                     folder_uri, gtk_get_current_event_time (), NULL);
+    MidoriBrowser* browser = midori_browser_get_for_widget (addons->treeview);
+    gboolean handled = FALSE;
+    g_signal_emit_by_name (midori_browser_get_current_tab (browser), "open-uri", folder_uri, &handled);
     g_free (folder_uri);
 }
 
@@ -618,7 +618,6 @@ addons_get_toolbar (MidoriViewable* viewable)
     if (!ADDONS (viewable)->toolbar)
     {
         toolbar = gtk_toolbar_new ();
-        gtk_toolbar_set_icon_size (GTK_TOOLBAR (toolbar), GTK_ICON_SIZE_BUTTON);
         toolitem = gtk_tool_item_new ();
         gtk_toolbar_insert (GTK_TOOLBAR (toolbar), toolitem, -1);
         gtk_widget_show (GTK_WIDGET (toolitem));
@@ -930,10 +929,16 @@ js_metadata_from_file (const gchar* filename,
                 /* We don't support these, so abort here */
                 g_free (line);
                 g_io_channel_shutdown (channel, false, 0);
-                g_slist_free (*includes);
-                g_slist_free (*excludes);
-                *includes = NULL;
-                *excludes = NULL;
+                if (includes != NULL)
+                {
+                    g_slist_free (*includes);
+                    *includes = NULL;
+                }
+                if (excludes != NULL)
+                {
+                    g_slist_free (*excludes);
+                    *excludes = NULL;
+                }
                 return FALSE;
             }
              else if (includes && g_str_has_prefix (line, "// @include"))
