@@ -21,15 +21,10 @@
 #include <glib/gi18n.h>
 #include <libsoup/soup.h>
 
-#if WEBKIT_CHECK_VERSION (1, 3, 11)
     #define LIBSOUP_USE_UNSTABLE_REQUEST_API
     #include <libsoup/soup-cache.h>
-#endif
 
 #include <config.h>
-#if HAVE_LIBNOTIFY
-    #include <libnotify/notify.h>
-#endif
 
 struct _MidoriPreferences
 {
@@ -243,6 +238,17 @@ midori_preferences_get_spell_languages (void)
 }
 #endif
 
+#ifdef G_OS_WIN32
+static void
+midori_preferences_theme_changed_cb (GtkWidget* button,
+                                     gpointer   user_data)
+{
+    MidoriSettings* settings = user_data;
+    midori_settings_set_theme_name (settings,
+        gtk_combo_box_text_get_active_text (GTK_COMBO_BOX_TEXT (button)));
+}
+#endif
+
 /**
  * midori_preferences_set_settings:
  * @settings: the settings
@@ -298,7 +304,8 @@ midori_preferences_set_settings (MidoriPreferences* preferences,
     {
     PAGE_NEW (GTK_STOCK_HOME, _("Startup"));
     FRAME_NEW (NULL);
-    label = katze_property_label (settings, "load-on-startup");
+    label = gtk_label_new (_("When Midori starts:"));
+    gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
     INDENTED_ADD (label);
     button = katze_property_proxy (settings, "load-on-startup", NULL);
     SPANNED_ADD (button);
@@ -307,7 +314,7 @@ midori_preferences_set_settings (MidoriPreferences* preferences,
     INDENTED_ADD (label);
     entry = katze_property_proxy (settings, "homepage", "address");
     SPANNED_ADD (entry);
-    if (parent && katze_object_has_property (parent, "uri"))
+    if (parent && MIDORI_IS_BROWSER (parent))
     {
         #if 0
         button = gtk_button_new_with_mnemonic (_("Use _current page"));
@@ -354,7 +361,7 @@ midori_preferences_set_settings (MidoriPreferences* preferences,
     label = gtk_label_new (_("Preferred Encoding"));
     gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
     INDENTED_ADD (label);
-    button = katze_property_proxy (settings, "preferred-encoding", "custom-default-encoding");
+    button = katze_property_proxy (settings, "preferred-encoding", "custom-default-charset");
     SPANNED_ADD (button);
 
     /* Page "Behavior" */
@@ -377,16 +384,24 @@ midori_preferences_set_settings (MidoriPreferences* preferences,
         else
             g_free (enchant_path);
     }
-    button = katze_property_proxy (settings, "enable-scripts", NULL);
+    button = katze_property_proxy (settings, "enable-javascript", NULL);
     gtk_button_set_label (GTK_BUTTON (button), _("Enable scripts"));
     INDENTED_ADD (button);
-    #if WEBKIT_CHECK_VERSION (1, 3, 8)
     button = katze_property_proxy (settings, "enable-webgl", NULL);
     gtk_button_set_label (GTK_BUTTON (button), _("Enable WebGL support"));
-    #else
-    button = katze_property_proxy (settings, "enable-plugins", NULL);
-    gtk_button_set_label (GTK_BUTTON (button), _("Enable Netscape plugins"));
-    gtk_widget_set_sensitive (button, midori_web_settings_has_plugin_support ());
+    #ifndef HAVE_WEBKIT2
+    if (parent && MIDORI_IS_BROWSER (parent))
+    {
+        GtkWidget* tab = midori_browser_get_current_tab (MIDORI_BROWSER (parent));
+        WebKitWebView* web_view = midori_tab_get_web_view (MIDORI_TAB (tab));
+        WebKitWebFrame* web_frame = webkit_web_view_get_main_frame (WEBKIT_WEB_VIEW (web_view));
+        JSContextRef js_context = webkit_web_frame_get_global_context (web_frame);
+        gchar* supports_web_gl = sokoke_js_script_eval (js_context,
+            "!!window.WebGLRenderingContext", NULL);
+        if (g_strcmp0 (supports_web_gl, "true"))
+            gtk_widget_hide (button);
+        g_free (supports_web_gl);
+    }
     #endif
     SPANNED_ADD (button);
     button = katze_property_proxy (settings, "zoom-text-and-images", NULL);
@@ -396,25 +411,12 @@ midori_preferences_set_settings (MidoriPreferences* preferences,
     gtk_button_set_label (GTK_BUTTON (button), _("Allow scripts to open popups"));
     gtk_widget_set_tooltip_text (button, _("Whether scripts are allowed to open popup windows automatically"));
     SPANNED_ADD (button);
-    if (katze_widget_has_touchscreen_mode (parent ?
-        GTK_WIDGET (parent) : GTK_WIDGET (preferences)))
-    {
-        button = katze_property_proxy (settings, "kinetic-scrolling", NULL);
-        gtk_button_set_label (GTK_BUTTON (button), _("Kinetic scrolling"));
-        gtk_widget_set_tooltip_text (button, _("Whether scrolling should kinetically move according to speed"));
-    }
-    else
-    {
-        button = katze_property_proxy (settings, "middle-click-opens-selection", NULL);
-        gtk_button_set_label (GTK_BUTTON (button), _("Middle click opens Selection"));
-        gtk_widget_set_tooltip_text (button, _("Load an address from the selection via middle click"));
-    }
-    INDENTED_ADD (button);
-    #ifndef G_OS_WIN32
-    button = katze_property_proxy (settings, "flash-window-on-new-bg-tabs", NULL);
-    gtk_button_set_label (GTK_BUTTON (button), _("Flash window on background tabs"));
+    label = gtk_label_new (_("Default Zoom Level"));
+    gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
+    INDENTED_ADD (label);
+    button = katze_property_proxy (settings, "zoom-level", NULL);
+    gtk_widget_set_tooltip_text (button, _("Initial factor to enlarge newly opened tabs by"));
     SPANNED_ADD (button);
-    #endif
 
     FRAME_NEW (NULL);
     button = gtk_label_new (_("Preferred languages"));
@@ -431,26 +433,55 @@ midori_preferences_set_settings (MidoriPreferences* preferences,
 
     /* Page "Interface" */
     PAGE_NEW (GTK_STOCK_CONVERT, _("Browsing"));
-    if (!g_getenv ("DESKTOP_SESSION"))
-    {
         FRAME_NEW (NULL);
-        INDENTED_ADD (katze_property_label (settings, "toolbar-style"));
+    #ifdef G_OS_WIN32
+    INDENTED_ADD (gtk_label_new (_("Theme:")));
+    button = gtk_combo_box_text_new ();
+    SPANNED_ADD (button);
+    guint i = 0;
+    /* On Windows the default theme may be a built-in specific to the
+       running system. So we always add the default and skip it later.
+     */
+    const gchar* default_theme = midori_settings_get_default_theme_name (MIDORI_SETTINGS (settings));
+    gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (button), default_theme);
+    gtk_combo_box_set_active (GTK_COMBO_BOX (button), i++);
+
+    const gchar* current_theme = midori_settings_get_theme_name (MIDORI_SETTINGS (settings));
+    gchar* theme_path = midori_paths_get_data_filename ("themes", FALSE);
+    GDir* dir;
+    if ((dir = g_dir_open (theme_path, 0, NULL)))
+    {
+        const gchar* name;
+        while ((name = g_dir_read_name (dir)))
+        {
+            if (!g_strcmp0 (name, default_theme))
+                continue;
+            gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (button), name);
+            if (!g_strcmp0 (name, current_theme))
+                gtk_combo_box_set_active (GTK_COMBO_BOX (button), i);
+            i++;
+        }
+        g_dir_close (dir);
+        g_signal_connect (button, "changed",
+                          G_CALLBACK (midori_preferences_theme_changed_cb), settings);
+    }
+    g_free (theme_path);
+    #endif
+        INDENTED_ADD (gtk_label_new (_("Toolbar Style:")));
         button = katze_property_proxy (settings, "toolbar-style", NULL);
         SPANNED_ADD (button);
-    }
     FRAME_NEW (NULL);
-    label = katze_property_label (settings, "open-new-pages-in");
+    label = gtk_label_new (_("Open new pages in:"));
     INDENTED_ADD (label);
     button = katze_property_proxy (settings, "open-new-pages-in", NULL);
     SPANNED_ADD (button);
+    label = gtk_label_new (_("New tab behavior:"));
+    INDENTED_ADD (label);
+    button = katze_property_proxy (settings, "new-tab-type", "custom-tabhome");
+    SPANNED_ADD (button);
     button = katze_property_proxy (settings, "close-buttons-on-tabs", NULL);
     gtk_button_set_label (GTK_BUTTON (button), _("Close Buttons on Tabs"));
     INDENTED_ADD (button);
-    #ifndef HAVE_GRANITE
-    button = katze_property_proxy (settings, "always-show-tabbar", NULL);
-    gtk_button_set_label (GTK_BUTTON (button), _("Always Show Tabbar"));
-    SPANNED_ADD (button);
-    #endif
     button = katze_property_proxy (settings, "open-tabs-next-to-current", NULL);
     gtk_button_set_label (GTK_BUTTON (button), _("Open Tabs next to Current"));
     gtk_widget_set_tooltip_text (button, _("Whether to open new tabs next to the current tab or after the last one"));
@@ -459,18 +490,6 @@ midori_preferences_set_settings (MidoriPreferences* preferences,
     gtk_button_set_label (GTK_BUTTON (button), _("Open tabs in the background"));
     SPANNED_ADD (button);
 
-    INDENTED_ADD (gtk_label_new (NULL));
-    label = gtk_label_new (_("Text Editor"));
-    gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
-    INDENTED_ADD (label);
-    entry = katze_property_proxy (settings, "text-editor", "application-text/plain");
-    SPANNED_ADD (entry);
-    label = gtk_label_new (_("News Aggregator"));
-    gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
-    INDENTED_ADD (label);
-    entry = katze_property_proxy (settings, "news-aggregator", "application-News");
-    SPANNED_ADD (entry);
-
     /* Page "Network" */
     PAGE_NEW (GTK_STOCK_NETWORK, _("Network"));
     FRAME_NEW (NULL);
@@ -479,37 +498,59 @@ midori_preferences_set_settings (MidoriPreferences* preferences,
     INDENTED_ADD (label);
     button = katze_property_proxy (settings, "proxy-type", NULL);
     SPANNED_ADD (button);
-    label = gtk_label_new (_("Hostname"));
+    label = gtk_label_new (_("URI"));
     gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
     INDENTED_ADD (label);
     entry = katze_property_proxy (settings, "http-proxy", "address");
     SPANNED_ADD (entry);
-    g_signal_connect (settings, "notify::proxy-type",
-        G_CALLBACK (midori_preferences_notify_proxy_type_cb), entry);
+    g_signal_connect_object (settings, "notify::proxy-type",
+        G_CALLBACK (midori_preferences_notify_proxy_type_cb), entry, 0);
     midori_preferences_notify_proxy_type_cb (settings, NULL, entry);
     label = gtk_label_new (_("Port"));
     gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
     INDENTED_ADD (label);
     entry = katze_property_proxy (settings, "http-proxy-port", NULL);
     SPANNED_ADD (entry);
-    g_signal_connect (settings, "notify::proxy-type",
-        G_CALLBACK (midori_preferences_notify_proxy_type_cb), entry);
+    g_signal_connect_object (settings, "notify::proxy-type",
+        G_CALLBACK (midori_preferences_notify_proxy_type_cb), entry, 0);
     midori_preferences_notify_proxy_type_cb (settings, NULL, entry);
-    #if WEBKIT_CHECK_VERSION (1, 3, 11)
+    INDENTED_ADD (gtk_event_box_new ());
+    label = gtk_label_new (NULL);
+    GString* proxy_types = g_string_new (NULL);
+    g_string_append (proxy_types, "<span size=\"smaller\">");
+    g_string_append (proxy_types, _("Supported proxy types:"));
+    g_string_append_c (proxy_types, ' ');
+    /* http is not implemented as a GIO extension */
+    g_string_append (proxy_types, "http:// https://");
+    GIOExtensionPoint* proxy_extension = g_io_extension_point_lookup ("gio-proxy");
+    GList* proxies = g_io_extension_point_get_extensions (proxy_extension);
+    for (; proxies; proxies = g_list_next (proxies))
+        g_string_append_printf (proxy_types, " %s://", g_io_extension_get_name (proxies->data));
+    g_string_append (proxy_types, "</span>");
+    gtk_label_set_markup (GTK_LABEL (label), proxy_types->str);
+    g_string_free (proxy_types, TRUE);
+    SPANNED_ADD (label);
+    g_signal_connect_object (settings, "notify::proxy-type",
+        G_CALLBACK (midori_preferences_notify_proxy_type_cb), label, 0);
+    midori_preferences_notify_proxy_type_cb (settings, NULL, label);
+#ifndef HAVE_WEBKIT2
     if (soup_session_get_feature (webkit_get_default_session (), SOUP_TYPE_CACHE))
     {
         label = gtk_label_new (_("Web Cache"));
         gtk_widget_set_tooltip_text (label, _("The maximum size of cached pages on disk"));
         INDENTED_ADD (label);
         button = katze_property_proxy (settings, "maximum-cache-size", NULL);
+        gtk_spin_button_set_range (GTK_SPIN_BUTTON (button), 0, G_MAXINT);
         gtk_widget_set_tooltip_text (button, _("The maximum size of cached pages on disk"));
         SPANNED_ADD (button);
         label = gtk_label_new (_("MB"));
         gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
         SPANNED_ADD (label);
     }
-    #endif
-    label = katze_property_label (settings, "identify-as");
+#endif
+    /* i18n: This refers to an application, not the 'user agent' string */
+    label = gtk_label_new (_("Identify as"));
+    gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
     INDENTED_ADD (label);
     button = katze_property_proxy (settings, "identify-as", "custom-user-agent");
     SPANNED_ADD (button);
@@ -523,6 +564,9 @@ midori_preferences_add_privacy_category (KatzePreferences*  preferences,
     GtkWidget* label;
     gchar* markup;
 
+    g_return_if_fail (KATZE_IS_PREFERENCES (preferences));
+    g_return_if_fail (MIDORI_IS_WEB_SETTINGS (settings));
+
     katze_preferences_add_category (preferences, _("Privacy"), GTK_STOCK_INDEX);
     katze_preferences_add_group (preferences, NULL);
     button = gtk_label_new (_("Delete old Cookies after:"));
@@ -573,7 +617,6 @@ midori_preferences_add_extension_category (KatzePreferences*  preferences,
                                            MidoriApp*         app)
 {
     KatzeArray* array;
-    gchar* extension_path;
     GtkWidget* scrolled;
     GtkWidget* addon;
     GList* children;
@@ -593,9 +636,9 @@ midori_preferences_add_extension_category (KatzePreferences*  preferences,
     }
     g_object_unref (array);
 
-    scrolled = g_object_new (KATZE_TYPE_SCROLLED, "visible", TRUE, NULL);
+    scrolled = gtk_scrolled_window_new (NULL, NULL);
     /* For lack of a better way of keeping descriptions visible */
-    g_object_set (scrolled, "hscrollbar-policy", GTK_POLICY_NEVER, NULL);
+    g_object_set (scrolled, "visible", TRUE, "hscrollbar-policy", GTK_POLICY_NEVER, NULL);
     addon = g_object_new (MIDORI_TYPE_EXTENSIONS, "app", app, NULL);
     children = gtk_container_get_children (GTK_CONTAINER (addon));
     gtk_widget_reparent (g_list_nth_data (children, 0), scrolled);
